#cs
(module lexer mzscheme
  (require (lib "lex.ss" "parser-tools"))
  (provide (all-defined-except get-token-name))
  
  (define-syntax define-open-tokens
    (syntax-rules ()
      ((_ dt group-name list-name (token-symbols ...))
       (dt group-name (token-symbols ...)))))
  
  (define-open-tokens define-empty-tokens Operators operator-tokens
    (= > < + - * /))

  (define-open-tokens define-empty-tokens Separators separator-tokens
   (O_PAREN C_PAREN O_BRACE C_BRACE O_BRACKET C_BRACKET SEMI_COLON COMMA DOUBLE_POINT AT COMMENT))
  
  (define-open-tokens define-empty-tokens EmptyLiterals literal-tokens
   (NULL_LIT TRUE_LIT FALSE_LIT))

  (define-open-tokens define-tokens pico-vals val-tokens
    (STRING_LIT CHAR_LIT DECIMAL_LIT IDENTIFIER))

  (define-lex-abbrevs
   [any-string (&)]
    (CR #\015)
    (LF #\012)
    (LineTerminator (: CR 
		       LF 
		       (@ CR LF)))
    (InputCharacter (^ CR LF))
    
    (FF #\014)
    (TAB #\011)
    (WhiteSpace (: #\space 
		   TAB
		   FF
		   LineTerminator))

    (sPicoSign (: "!" "%" "&" "#" "?" "~" "^" "$" SignOperator))
    (Identifier (: (+ sPicoSign) (@ sPicoLetter (* sPicoLetterOrDigit))))
    (sPicoLetter (: (- "A" "Z")
		                (- "a" "z")
		                "_"))
    
    (Digit (- #\0 #\9))
    (Exponent (@ "e" (? "+") (* Digit)))
    (DecimalNumber (@ (+ Digit) (? (@ #\. (* Digit))) (? Exponent)))
    
    (sPicoLetterOrDigit (: sPicoLetter Digit))

    (Keyword (: "accept" "eval" "call" "void" "eoln" "display" "trunc" "random" "number"
                "text" "abs" "ord" "sqrt" "sin" "cos" "tan" "arcsin" "arccos" "arctan" "exp"
		            "log" "explode" "implode" "is_number" "is_fraction" "is_text" "is_function"
		            "is_table" "is_void" "size" "tab" "and" "or" "not" "begin" "continue" "if"
		            "while" "until" "for" "load" "read" "print" "tag" "make" "interrupt" "error"))
    
    (SignOperator (: "=" "+" "-" "*" "/" "^" "<" ">" ))
    (Operator (: SignOperator "equivalent" )))
 
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;;Comment lexer
  (define read-block-comment
    (lexer
     ["`" end-pos]
     [(eof) end-pos]
     [(: "`" (~ (@ any-string (: "`") any-string))) (read-block-comment input-port)]))
  
  ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
  ;String lexer
  
  (define (get-string input-port)
    (letrec ((tokens (get-string-tokens input-port))
             (last-token (list-ref tokens (sub1 (length tokens))))
             (tokens->string
              (lambda (toks)
                (if (null? (cdr toks))
                    ""
                    (string-append (string (token-value (car (car toks))))
                                   (tokens->string (cdr toks)))))))
      (if (eq? 'STRING_END (get-token-name last-token))
          (token-STRING_LIT (list (tokens->string tokens) (caddr last-token))))))
  
  (define (get-string-tokens input-port)
    (let ((tok (get-str-tok input-port)))
      (case (get-token-name tok)
        ((STRING_EOF STRING_END STRING_NEWLINE) (list tok))
        (else (cons tok (get-string-tokens input-port))))))

  (define (get-token-name tok)
    (if (token? (car tok))
        (token-name (car tok))
        (car tok)))
  
  (define-tokens str-tok (STRING_CHAR))
  (define-empty-tokens err (STRING_END STRING_EOF STRING_NEWLINE))
  
  (define get-str-tok
    (lexer-src-pos
     (#\" (token-STRING_END))
     ((^ CR LF) (token-STRING_CHAR (string-ref lexeme 0)))
     ((: CR LF) (token-STRING_NEWLINE))
     (#\032 (token-STRING_EOF))
     ((eof) (token-STRING_EOF))))
	    
  (define get-token
    (lexer-src-pos
     (Operator (let ((l lexeme))
                 (string->symbol l)))
     ("(" (token-O_PAREN))
     (")" (token-C_PAREN))
     ("{" (token-O_BRACE))
     ("}" (token-C_BRACE))
     ("[" (token-O_BRACKET))
     ("]" (token-C_BRACKET))
     (";" (token-SEMI_COLON))
     ("," (token-COMMA))
     (":" (token-DOUBLE_POINT))
     ("@" (token-AT))

     (#\" (get-string input-port))
     (#\' (get-string input-port))

     ("true" (token-TRUE_LIT))
     ("false" (token-FALSE_LIT))

     (Keyword (string->symbol lexeme))
     (Identifier (token-IDENTIFIER lexeme))
     (DecimalNumber (token-DECIMAL_LIT (string->number lexeme)))
     ((+ WhiteSpace) (return-without-pos (get-token input-port)))
     (#\032 'EOF)
     ((eof) 'EOF)
     ))
  
  (define (syn-val lex a b c d)
    (values lex a b (position-offset c) (position-offset d)))
  
  (define get-syn-string
    (lexer
     ((: CR LF #\") (position-offset end-pos))
     ((eof) (position-offset end-pos))
     ((^ CR LF) (get-syn-string input-port))))
  
  (define colorize-string1
    (lexer
     ["\"" end-pos]
     [(eof) end-pos]
     [(: "\"" (~ (@ any-string (: "\"") any-string))) (colorize-string1 input-port)]))
 
  (define colorize-string2
    (lexer
     ["'" end-pos]
     [(eof) end-pos]
     [(: "'" (~ (@ any-string (: "'") any-string))) (colorize-string2 input-port)]))
  
  (define get-syntax-token
    (lexer
     (Operator
      (syn-val lexeme 'keyword #f start-pos end-pos))
     
     ((: "null" "true" "false" DecimalNumber)
      (syn-val lexeme 'literal #f start-pos end-pos))

     ((: "(" ")" "{" "}" "[" "]")
      (syn-val lexeme 'seperator (string->symbol lexeme) start-pos end-pos))
     ((: ";" "," ":" "@")
      (syn-val lexeme 'seperator #f start-pos end-pos))
    
     ("\"" (syn-val lexeme 'literal #f start-pos (colorize-string1 input-port)))
     ("'" (syn-val lexeme 'literal #f start-pos (colorize-string2 input-port)))

     (Keyword (syn-val lexeme 'keyword #f start-pos end-pos))

     (Identifier (syn-val lexeme 'identifier #f start-pos end-pos))

     ("`" (syn-val lexeme 'comment #f start-pos (read-block-comment input-port)))

     ((+ WhiteSpace) (syn-val lexeme 'white-space #f start-pos end-pos))

     (#\032 (values lexeme 'eof #f start-pos end-pos))
     ((eof) (values lexeme 'eof #f start-pos end-pos))

     ((+ (: (- #\0 #\9)(- #\a #\z)(- #\A #\Z))) (syn-val lexeme 'error #f start-pos end-pos))
     
     ((- #\000 #\377) (syn-val lexeme 'error #f start-pos end-pos))
     
     ))
  )
